#include "../../misc/log_time.h"

#include "xmlwrap.h"
#include "service.h"

#include "object.h"


// *** Register objects ***

extern FlOObject* deserialize_knot(const XMLNode& x);
class FlOObjectRegistrator {
 public:
  FlOObjectRegistrator(void) {
    FlOObject::Register(&deserialize_knot);
  };
};

std::list<FlOObject::deserializer_t> FlOObject::deserializers;
static FlOObjectRegistrator registrator;


// *** FlOObject ***

void FlOObject::Register(FlOObject::deserializer_t sr) {
  deserializers.push_back(sr);
}

FlOObject::FlOObject(const std::string& id,type_t t):service_(NULL),id_(id),type_(t) {
  odlog(3)<<"FlOObject: constructor: id: "<<id_<<endl;
  odlog(3)<<"FlOObject: constructor: type: "<<type_<<endl;
}

FlOObject::FlOObject(FlOService& service,const std::string& id,type_t t):service_(&service),id_(id),type_(t) {
  odlog(3)<<"FlOObject: constructor: id: "<<id_<<endl;
  odlog(3)<<"FlOObject: constructor: type: "<<type_<<endl;
  pFlOObject p(*this);
  // Add itself to service
  odlog(3)<<"FlOObject: constructor: adding to service"<<endl;
  service.add(p);
  // // Notify registered services
  // odlog(3)<<"FlOObject: constructor: notifying neighbours"<<endl;
  // service.notify(FlOMessage(*this,created));
}

FlOObject::FlOObject(const FlOObject& o):service_(NULL) {
  type_=o.type_;
  id_=o.id_;
  for(std::vector<FlOAttribute>::const_iterator i = o.attrs_.begin();
                i!=o.attrs_.end();++i) {
    attrs_.push_back(*i);
  };
}

FlOObject::~FlOObject(void) {
  odlog(3)<<"FlOObject: destructor"<<endl;
}
 
void FlOObject::serialize(std::string& msg) const {
  odlog(3)<<"FlOObject: serialize to string"<<endl;
  XMLNode x;
  serialize(x);
  msg=x.get();
}

void FlOObject::serialize(XMLNode& x) const {
  odlog(3)<<"FlOObject: serialize to xml"<<endl;
  x.name("object");
  XMLNode x_id(x,"id");
  XMLNode x_type(x,"type");
  switch(type_) {
    case FlOObject::reference: x_type.content("reference"); break;
    case FlOObject::ordinary:  x_type.content("ordinary");  break;
  };
  x_id.content(id_);
  for(vector<FlOAttribute>::const_iterator i = attrs_.begin();
                                                   i!=attrs_.end();++i) {
    XMLNode x_attr(x,"attribute");
    i->serialize(x_attr);
  };
  odlog(3)<<"FlOObject: serialize: "<<x.get()<<endl;
}

FlOObject* FlOObject::deserialize(const std::string& msg) {
  XMLNode x;
  x.set(msg);
  return deserialize(x);
}

FlOObject* FlOObject::deserialize(const XMLNode& x) {
  if(!x) return NULL;
  FlOObject* o = NULL;
  for(std::list<FlOObject::deserializer_t>::const_iterator d =
            deserializers.begin();d!=deserializers.end();++d) {
    o = (*(*d))(x);
    if(o == NULL) continue;
    if(!(*o)) { delete o; return NULL; }; // object was recognized and failed
  };
  return o;
};

FlOObject* deserialize_object(const std::string& msg) {
  XMLNode x;
  x.set(msg);
  return deserialize_object(x);
}

FlOObject* deserialize_object(const XMLNode& x) {
  if(x.name() != "object") return NULL;
  FlOObject* o = NULL;
  XMLNode* x_id = x.getNode("id");
  XMLNode* x_type = x.getNode("type");
  if((x_id != NULL) && (x_type != NULL)) {
    FlOObject::type_t o_type = FlOObject::undefined;
    std::string cnt;
    cnt=x_type->content();
    if(cnt == "reference") { o_type=FlOObject::reference; }
    else if(cnt == "ordinary") { o_type=FlOObject::ordinary; };
    if(o_type != FlOObject::undefined) {
      o=new FlOObject(x_id->content(),o_type);
    };
  };
  if(x_id != NULL) delete x_id;
  if(x_type != NULL) delete x_type;
  if(o) {
    for(int n = 0;;n++) {
      XMLNode* x_attr = x.getNode("attribute",n);
      if(!x_attr) break;
      FlOAttribute* attr = FlOAttribute::deserialize(*x_attr);
      if(!attr) { delete x_attr; continue; };
      o->attrs_.push_back(*attr);
      delete x_attr;
      delete attr;
    };
  };
  return o;
}

void FlOObject::serialize(XMLNode& x,FlOMessage& m) const {
  odlog(3)<<"FlOObject: serialize for message to xml"<<endl;
  serialize(x);
}

void FlOObject::serialize(std::string& msg,FlOMessage& m) const {
  odlog(3)<<"FlOObject: serialize for message to string"<<endl;
  serialize(msg);
}

FlOObject* FlOObject::deserialize(const XMLNode& x,FlOMessage& m) {
  odlog(3)<<"FlOObject: deserialize for message"<<endl;
  return deserialize(x);
}

FlOObject* FlOObject::deserialize(const std::string& msg,FlOMessage& m) {
  odlog(3)<<"FlOObject: deserialize for message"<<endl;
  return deserialize(msg);
}

FlOObject::operator bool(void) {
  odlog(3)<<"FlOObject: bool: "<<(type_ != undefined)<<endl;
  return (type_ != undefined);
}

bool FlOObject::operator!(void) {
  odlog(3)<<"FlOObject: !: "<<(type_ == undefined)<<endl;
  return (type_ == undefined);
}

void FlOObject::maintain(void) {
  odlog(3)<<"FlOObject: maintain"<<endl;


}


void FlOObject::notification(FlOMessage& msg,pFlOServiceRemote rsrv) {
  odlog(3)<<"FlOObject: notification"<<endl;
  if(type_ == reference) {
    odlog(3)<<"FlOObject: notification: reference"<<endl;
    switch(msg.type()) {
      case created: {
        odlog(3)<<"FlOObject: notification: created"<<endl;
        // Check for race condition/duplicate id (TODO)


        // We already know about creation of this object.
        // Do not propagate information through subscribed services
        // to avoid loops. TODO: do that in a more sophisticated way (timestamp)
      }; break;
      case updated: {
        odlog(3)<<"FlOObject: notification: updated"<<endl;
        // Avoid propagation loops 

        // Propagate notification to linked objects
        for(std::vector<pFlOServiceRemote>::iterator r = links_.begin();
                                                       r!=links_.end();++r) {
          (*r)->notify(msg);
        };
      }; break;
      case deleted: {
        odlog(3)<<"FlOObject: notification: deleted"<<endl;
        // Propagate notification to linked objects
        for(std::vector<pFlOServiceRemote>::iterator r = links_.begin();
                                                       r!=links_.end();++r) {
          (*r)->notify(msg);
        };
      }; break;
    };
  } else if(type_ = ordinary) { // Not implemented yet
    odlog(3)<<"FlOObject: notification: ordinary"<<endl;
    switch(msg.type()) {
      case created: {
        odlog(3)<<"FlOObject: notification: created"<<endl;
      }; break;
      case updated: {
        odlog(3)<<"FlOObject: notification: updated"<<endl;
      }; break;
      case deleted: {
        odlog(3)<<"FlOObject: notification: deleted"<<endl;
      }; break;
    };
  };
}

// This is an object belonging to message
void FlOObject::notification(FlOService& srv,FlOMessage& msg,pFlOServiceRemote rsrv) const {
  odlog(3)<<"FlOObject: static notification"<<endl;
  switch(msg.type()) {
    case created:
    case updated: {
      odlog(3)<<"FlOObject: static notification: created/updated"<<endl;
      // Decide if this object should be stored here

      // Create new object (if needed) which belongs to calling service
      // So far only reference objects are created on demand
      pFlOObject obj = new FlOObject(srv,msg.object()->id(),reference);
      // Add link to server which sent notification
      obj->links_.push_back(rsrv);
      // Add object to service
      srv.add(obj);
      // Pass notification to services subscribed for general notifications
      // through local service (it knows what to do)
      srv.notify(msg);
      return;
    }; break;
    case deleted: {
      odlog(3)<<"FlOObject: static notification: deleted"<<endl;
      // Deleted and never existed
    }; break;
  };
  return;
}

