package se.umu.jobsubmission.impl.informationfinder; 

import java.net.URI; 

import java.util.HashSet; 
import java.util.Iterator; 
import java.util.List; 
import java.util.LinkedList;
import java.util.Set; 
import java.util.Vector; 

import org.apache.log4j.Logger;

import se.umu.jobsubmission.util.ThreadPool; 
import se.umu.jobsubmission.util.Worker; 

import org.globus.mds.glue.GLUECERPType;

import se.umu.jobsubmission.impl.informationfinder.ResourceFinder; 
import se.umu.jobsubmission.impl.informationfinder.InformationFetcher; 
import se.umu.jobsubmission.impl.informationfinder.InformationConverter; 

public class InformationFinder {

    // Threadpool
    private int noThreads; 
    private ThreadPool pool; 

    // Which finder, fetcher and converter to use...
    private String finderClasses[]; 
    private String fetcherClasses[]; 
    private String converterClasses[]; 
    private int timeout; // timeout for finder and fetcher. 

    private List indexServers = null; // default index server(s) to contact 
                       // (in case user does not specify any) 
    
    private ResourceInformationCache resourceInformationCache; 
    
    private static Logger logger = Logger.getLogger(InformationFinder.class); 

    public InformationFinder(int noThreads, 
            String finderClasses[], 
            String fetcherClasses[], 
            String converterClasses[], 
            int timeout, 
            List indexServers, 
			ResourceInformationCache resourceInformationCache) {
        this.noThreads = noThreads; 
        this.finderClasses = finderClasses; 
        this.fetcherClasses = fetcherClasses; 
        this.converterClasses = converterClasses; 
        this.timeout = timeout; 
        this.indexServers = indexServers; 
        this.resourceInformationCache = resourceInformationCache; 

        pool = new ThreadPool(noThreads); 

        logger.debug("InformationFinder started"); 
        for(int i = 0; i < finderClasses.length; i++) {
        	logger.debug("Finder - " + finderClasses[i]);
        }
        for(int i = 0; i < fetcherClasses.length; i++) {
        	logger.debug("Fetcher - " + fetcherClasses[i]); 
        }
        if(this.converterClasses != null) {
        	for(int i = 0; i < converterClasses.length; i++) {
        		logger.debug("Converter - " + converterClasses[i]);
        	}
        }
        logger.debug("timeout - " + timeout); 
        logger.debug("#Threads - " + noThreads); 
        for(int i = 0; i < indexServers.size(); i++)
            logger.debug("Default index server: " + indexServers.get(i)); 
    }

    // Find available resources: 
    public List findResources(List indexServerURIs) throws Exception {
 
        List resourceURIs = null; 

        try {
            logger.info("Starting with " + 
                    indexServerURIs.size() + " index servers"); 

            // Start initial worker(s)
            List initialWorkers = new LinkedList(); 
            Iterator argIter = indexServerURIs.iterator(); 
            while(argIter.hasNext()) {
                Worker worker = new FinderWorker(argIter.next(), 
                		// finder, 
                		this.finderClasses, 
						this.timeout);
                initialWorkers.add(worker); 
            }

            // Run workers in workpool, await until pool is done. 
            resourceURIs = pool.work(initialWorkers); 
            logger.info("Found " + resourceURIs.size() + " resources"); 
          
        } catch (Exception e) {
            logger.info("Error in findResources"); 
            return resourceURIs; 
        }
        
        if(resourceURIs.size() == 0) {
        	throw new Exception("No resources found"); 
        }
        return resourceURIs; 

    }

    // Query each resource: 
    public List fetchInformation(List resourceURIs) {

        List resourceInfo = null; 

        try {

            logger.info("Getting info from " + resourceURIs.size() 
                    + " resources"); 
            
            // Randomize list, to improve cache performance. 
            int size = resourceURIs.size(); 
            Object resourceArray[] = resourceURIs.toArray(); 
            for(int i = 0; i < size; i++) {
            	int index = (int) (Math.random()*size); 
            	Object tmp = resourceArray[i]; 
            	resourceArray[i] = resourceArray[index]; 
            	resourceArray[index] = tmp; 
            }            
            resourceURIs.clear(); 
            for(int i = 0; i < size; i++) {
            	resourceURIs.add(resourceArray[i]); 
            }

            // Create a new worker for each resource and add it to threadpool. 
            List initialWorkers = new LinkedList(); 
            Iterator argIter = resourceURIs.iterator(); 
            while(argIter.hasNext()) {
            	
                Worker worker = new FetcherWorker(argIter.next(), 
                		this.fetcherClasses, 
						this.timeout, 
						this.resourceInformationCache); 
                initialWorkers.add(worker); 
            }

            // Wait for results. 
            resourceInfo = pool.work(initialWorkers); 
            logger.info("Got information from " + resourceInfo.size()
                    + " resources"); 

            return resourceInfo; 
        } catch (Exception e) {
            logger.info("Error in fetchInformation"); 
            return resourceInfo; 
        }
    }

    // Convert all information, unless it already follows the Glue format. 
    // This is the case for LCG, where the informationsystem uses Glue. 
    public List convertInformation(List resourceInfo, List resourceURIs) 
    throws Exception {

        List GLUEInfo = null; 
        
        boolean allGlue = true; 
        for(int i = 0; i < resourceInfo.size(); i++) {
        	if(!(resourceInfo.get(i) instanceof GLUECERPType)) {
        		allGlue = false; 
        	}
        }
        if(allGlue && this.converterClasses == null) {
        	logger.info("Retrieved information already follows the glue format"); 
        	return resourceInfo; 
        }

        try {
        	
        	if(this.converterClasses == null) {
        		logger.error("No valid information converters specified");
        		throw new Exception("No valid information converters specified"); 
        	}

            logger.info("Converting information from " + resourceInfo.size() 
                    + " resources"); 

            List initialWorkers = new LinkedList(); 
            
            // TODO: What if resourceInfo and resourceURIs does not have
            // the same size? -> This should not happen...            
            
            for(int i = 0; i < resourceInfo.size(); i++) {
            	            	            	
            	// Extract hostname from resourceURI if it is an URL, 
            	// otherwise, assume that whole resourceURI is a hostname  
            	String hostName = (String) resourceURIs.get(i); 
            	// if hostname is a full URL, extract 'host' part of it
            	if(hostName.matches(":")) {
	            	try { 
	            		URI uri = new URI(hostName); 
	            		hostName = uri.getHost(); 
	            	} catch (Exception e) {
	            		hostName = (String) resourceURIs.get(i); 
	            	} 
            	}
            	
            	logger.debug("ResourceKey (Hostname) is: " + hostName); 
            	
            	// The resourceInfo list contains null entries for 
            	// non-responding resources...
            	if(resourceInfo.get(i) == null) {
            		logger.info("No info retrieved from resource: " + hostName 
            				+ " ignoring it");
            	} 
            	else {
            		Worker worker = new ConverterWorker(resourceInfo.get(i), 
            				this.converterClasses, 
							hostName, // resourceURIs.get(i), 
							this.resourceInformationCache); 
            		initialWorkers.add(worker);
            	}
            }

            GLUEInfo = pool.work(initialWorkers); 

            logger.info("Converters created " + 
            		(GLUEInfo == null ? 0 : GLUEInfo.size())  
					+ " GLUE objects"); 

            return GLUEInfo; 
        } catch (Exception e) {
            logger.info("Error in convertInformation"); 
            return GLUEInfo; 
        }
    }


    /**
     * Removes all duplicates of values from a list by inserting 
     * all items in a Set, and then constructing a new list from the set. 
     * @param list
     * @return
     */
    public List removeDuplicates(List list) {
    	
    	Set set = new HashSet(); 
    	for(int i = 0; i < list.size(); i++) {
    		set.add(list.get(i)); 
    	}
    	
    	List result = new LinkedList(); 
    	Iterator iter = set.iterator(); 
    	while(iter.hasNext()) {
    		result.add(iter.next()); 
    	}
	
    	return result;     	    	
    }

    // If indexServers is true, the first parameter contains a list 
    // of index server URLs. Otherwise, it contains URLs to single
    // resources. 
    public List getResourceInfo(List indexServerURIs, boolean useIndexServers)
    		throws Exception {

        // Perform resource discovery and fetch information
        if(useIndexServers) {
        	
            // If no index servers are specified, use default ones. 
        	if(indexServerURIs == null) {
        		indexServerURIs = this.indexServers; 
        	}
        	
        	if(indexServerURIs == null) {
        		throw new Exception("No index server(s) specified"); 
        	}
        	
        	// filter initial list for duplicates
        	List uniqueIndexServers = removeDuplicates(indexServerURIs);
        	List resourceURIs = findResources(uniqueIndexServers);         
        	this.pool.reset(); 
        	
        	List resourceInfo = fetchInformation(resourceURIs); 
        	this.pool.reset();
        	
        	List GLUEInfo = convertInformation(resourceInfo, resourceURIs);
        	this.pool.reset();              
        
        	return GLUEInfo; 
        	
        }
        // Do not perform discovery if we already specified 
        // the resources to query...
        else {
        	
        	List uniqueResourceURLs = removeDuplicates(indexServerURIs); 
        	List resourceInfo = fetchInformation(uniqueResourceURLs); 
        	this.pool.reset(); 
        	
        	List GLUEInfo = convertInformation(resourceInfo, uniqueResourceURLs);
        	this.pool.reset();        
        	
        	return GLUEInfo; 
        	
        }
    }

    public void shutdown() {
    	logger.debug("Shutting down"); 
        this.pool.shutdown(); 
    }
    
    public ResourceInformationCache getResourceInformationCache() {
    	return this.resourceInformationCache; 
    }
}
